/*
 * Copyright 2000-2021 JetBrains s.r.o.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.jetbrains.internal;

import jdk.internal.org.objectweb.asm.*;

import java.lang.invoke.MethodHandle;
import java.lang.invoke.MethodHandles;
import java.lang.invoke.MethodType;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.*;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.Supplier;

import static com.jetbrains.internal.ASMUtils.*;
import static java.lang.invoke.MethodHandles.Lookup;
import static jdk.internal.org.objectweb.asm.Opcodes.*;

/**
 * This class generates {@linkplain Proxy proxy} classes.
 * Each proxy is just a generated class implementing some interface and
 * delegating method calls to method handles.
 * <p>
 * There are 2 proxy dispatch modes:
 * <ul>
 *     <li>interface -> proxy -> {@linkplain #generateBridge bridge} -> method handle -> implementation code</li>
 *     <li>interface -> proxy -> method handle -> implementation code</li>
 * </ul>
 * Generated proxy is always located in the same package with its interface and optional bridge is located in the
 * same module with target implementation code. Bridge allows proxy to safely call hidden non-static implementation
 * methods and is only needed for {@code jetbrains.api} -> JBR calls. For JBR -> {@code jetbrains.api} calls, proxy can
 * invoke method handle directly.
 */
class ProxyGenerator {
    private static final String OBJECT_DESCRIPTOR = "Ljava/lang/Object;";
    private static final String MH_NAME = "java/lang/invoke/MethodHandle";
    private static final String MH_DESCRIPTOR = "Ljava/lang/invoke/MethodHandle;";
    private static final String CONVERSION_DESCRIPTOR = "(Ljava/lang/Object;)Ljava/lang/Object;";
    /**
     * Print warnings about usage of deprecated interfaces and methods to {@link System#err}.
     */
    private static final boolean LOG_DEPRECATED = System.getProperty("jetbrains.api.logDeprecated", "true").equalsIgnoreCase("true");

    private static final AtomicInteger nameCounter = new AtomicInteger();

    private final ProxyInfo info;
    private final boolean generateBridge;
    private final String proxyName, bridgeName;
    private final ClassVisitor proxyWriter, bridgeWriter;
    private final List<Supplier<MethodHandle>> handles = new ArrayList<>();
    private final List<Supplier<Class<?>>> classReferences = new ArrayList<>();
    private final Set<Proxy<?>> directProxyDependencies = new HashSet<>();
    private final List<Exception> exceptions = new ArrayList<>();
    private int bridgeMethodCounter;
    private boolean allMethodsImplemented = true;
    private Lookup generatedHandlesHolder, generatedProxy;

    /**
     * Creates new proxy generator from given {@link ProxyInfo},
     * looks for abstract interface methods, corresponding implementation methods
     * and generates proxy bytecode. However, it doesn't actually load generated
     * classes until {@link #defineClasses()} is called.
     */
    ProxyGenerator(ProxyInfo info) {
        this.info = info;
        generateBridge = info.type != ProxyInfo.Type.CLIENT_PROXY;
        int nameId = nameCounter.getAndIncrement();
        proxyName = Type.getInternalName(info.interFace) + "$$JBRApiProxy$" + nameId;
        bridgeName = generateBridge ? info.apiModule.lookupClass().getPackageName().replace('.', '/') + "/" +
                info.interFace.getSimpleName() + "$$JBRApiBridge$" + nameId : null;

        class ClassWriter extends jdk.internal.org.objectweb.asm.ClassWriter {
            ClassWriter() { super(ClassWriter.COMPUTE_FRAMES); }
            ClassVisitor createEmptyVisitor() {
                return new ClassVisitor(api) {
                    @Override
                    public MethodVisitor visitMethod(int access, String name, String descriptor, String signature, String[] exceptions) {
                        return new MethodVisitor(api) {};
                    }
                };
            }
        }
        ClassWriter proxyClassWriter = new ClassWriter();
        proxyWriter = proxyClassWriter;
        bridgeWriter = generateBridge ? new ClassWriter() : proxyClassWriter.createEmptyVisitor();
        proxyWriter.visit(CLASSFILE_VERSION, ACC_SUPER | ACC_FINAL | ACC_SYNTHETIC, proxyName, null,
                "java/lang/Object", new String[] {Type.getInternalName(info.interFace)});
        bridgeWriter.visit(CLASSFILE_VERSION, ACC_SUPER | ACC_FINAL | ACC_SYNTHETIC | ACC_PUBLIC, bridgeName, null,
                "java/lang/Object", null);
        generateConstructor();
        generateMethods();
    }

    boolean areAllMethodsImplemented() {
        return allMethodsImplemented;
    }

    Set<Proxy<?>> getDirectProxyDependencies() {
        return directProxyDependencies;
    }

    /**
     * Insert all method handles and class references into static fields, so that proxy can call implementation methods.
     */
    void init() {
        try {
            for (int i = 0; i < handles.size(); i++) {
                generatedHandlesHolder
                        .findStaticVarHandle(generatedHandlesHolder.lookupClass(), "h" + i, MethodHandle.class)
                        .set(handles.get(i).get());
            }
            for (int i = 0; i < classReferences.size(); i++) {
                generatedHandlesHolder
                        .findStaticVarHandle(generatedHandlesHolder.lookupClass(), "c" + i, Class.class)
                        .set(classReferences.get(i).get());
            }
        } catch (NoSuchFieldException | IllegalAccessException e) {
            throw new RuntimeException(e);
        }
    }

    Class<?> getProxyClass() {
        return generatedProxy.lookupClass();
    }

    /**
     * @return method handle to constructor of generated proxy class.
     * <ul>
     *     <li>For {@linkplain ProxyInfo.Type#SERVICE services}, constructor is no-arg.</li>
     *     <li>For non-{@linkplain ProxyInfo.Type#SERVICE services}, constructor is single-arg,
     *     expecting target object to which it would delegate method calls.</li>
     * </ul>
     */
    MethodHandle findConstructor() {
        try {
            if (info.target == null) {
                return generatedProxy.findConstructor(generatedProxy.lookupClass(), MethodType.methodType(void.class));
            } else {
                MethodHandle c = generatedProxy.findConstructor(generatedProxy.lookupClass(),
                        MethodType.methodType(void.class, Object.class));
                if (info.type == ProxyInfo.Type.SERVICE) {
                    try {
                        return MethodHandles.foldArguments(c, info.target.findConstructor(info.target.lookupClass(),
                                MethodType.methodType(void.class)).asType(MethodType.methodType(Object.class)));
                    } catch (NoSuchMethodException | IllegalAccessException e) {
                        throw new RuntimeException("Service implementation must have no-args constructor: " +
                                info.target.lookupClass(), e);
                    }
                }
                return c;
            }
        } catch (IllegalAccessException | NoSuchMethodException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * @return method handle that receives proxy and returns its target, or null
     */
    MethodHandle findTargetExtractor() {
        if (info.target == null) return null;
        try {
            return generatedProxy.findGetter(generatedProxy.lookupClass(), "target", Object.class);
        } catch (NoSuchFieldException | IllegalAccessException e) {
            throw new RuntimeException(e);
        }
    }

    /**
     * Loads generated classes.
     */
    void defineClasses() {
        try {
            Lookup bridge = !generateBridge ? null : MethodHandles.privateLookupIn(
                    info.apiModule.defineClass(((ClassWriter) bridgeWriter).toByteArray()), info.apiModule);
            generatedProxy = MethodHandles.privateLookupIn(
                    info.interFaceLookup.defineClass(((ClassWriter) proxyWriter).toByteArray()), info.interFaceLookup);
            generatedHandlesHolder = generateBridge ? bridge : generatedProxy;
        } catch (IllegalAccessException e) {
            throw new RuntimeException(e);
        }
    }

    private void generateConstructor() {
        if (info.target != null) {
            proxyWriter.visitField(ACC_PRIVATE | ACC_FINAL, "target", OBJECT_DESCRIPTOR, null, null);
        }
        MethodVisitor p = proxyWriter.visitMethod(ACC_PRIVATE, "<init>", "(" +
                (info.target == null ? "" : OBJECT_DESCRIPTOR) + ")V", null, null);
        if (LOG_DEPRECATED && info.interFace.isAnnotationPresent(Deprecated.class)) {
            logDeprecated(p, "Warning: using deprecated JBR API interface " + info.interFace.getName());
        }
        p.visitVarInsn(ALOAD, 0);
        if (info.target != null) {
            p.visitInsn(DUP);
            p.visitVarInsn(ALOAD, 1);
            p.visitFieldInsn(PUTFIELD, proxyName, "target", OBJECT_DESCRIPTOR);
        }
        p.visitMethodInsn(INVOKESPECIAL, "java/lang/Object", "<init>", "()V", false);
        p.visitInsn(RETURN);
        p.visitMaxs(-1, -1);
    }

    private void generateMethods() {
        for (Method method : info.interFace.getMethods()) {
            int mod = method.getModifiers();
            if (!Modifier.isAbstract(mod)) continue;
            MethodMapping methodMapping = getTargetMethodMapping(method);

            Exception e1 = null;
            if (info.target != null) {
                try {
                    MethodHandle handle = info.target.findVirtual(
                            info.target.lookupClass(), method.getName(), methodMapping.type);
                    generateMethod(method, handle, methodMapping, true);
                    continue;
                } catch (NoSuchMethodException | IllegalAccessException e) {
                    e1 = e;
                }
            }

            Exception e2 = null;
            ProxyInfo.StaticMethodMapping mapping = info.staticMethods.get(method.getName());
            if (mapping != null) {
                try {
                    MethodHandle staticHandle =
                            mapping.lookup.findStatic(mapping.lookup.lookupClass(), mapping.methodName, methodMapping.type);
                    generateMethod(method, staticHandle, methodMapping, false);
                    continue;
                } catch (NoSuchMethodException | IllegalAccessException e) {
                    e2 = e;
                }
            }

            if (e1 != null) exceptions.add(e1);
            if (e2 != null) exceptions.add(e2);
            generateUnsupportedMethod(proxyWriter, method);
            allMethodsImplemented = false;
        }
    }

    private void generateMethod(Method interfaceMethod, MethodHandle handle, MethodMapping mapping, boolean passInstance) {
        InternalMethodInfo methodInfo = getInternalMethodInfo(interfaceMethod);
        String bridgeMethodDescriptor = mapping.getBridgeDescriptor(passInstance);

        ClassVisitor handleWriter = generateBridge ? bridgeWriter : proxyWriter;
        String bridgeOrProxyName = generateBridge ? bridgeName : proxyName;
        String handleName = addHandle(handleWriter, () -> handle);
        for (TypeMapping m : mapping) {
            if (m.conversion == TypeConversion.EXTRACT_TARGET ||
                    m.conversion == TypeConversion.DYNAMIC_2_WAY) {
                Proxy<?> from = m.fromProxy;
                m.metadata.extractTargetHandle = addHandle(handleWriter, from::getTargetExtractor);
                directProxyDependencies.add(from);
            }
            if (m.conversion == TypeConversion.WRAP_INTO_PROXY ||
                    m.conversion == TypeConversion.DYNAMIC_2_WAY) {
                Proxy<?> to = m.toProxy;
                m.metadata.proxyConstructorHandle = addHandle(handleWriter, to::getConstructor);
                directProxyDependencies.add(to);
            }
            if (m.conversion == TypeConversion.DYNAMIC_2_WAY) {
                String classField = "c" + classReferences.size();
                m.metadata.extractableClassField = classField;
                classReferences.add(m.fromProxy::getProxyClass);
                handleWriter.visitField(ACC_PRIVATE | ACC_STATIC, classField, "Ljava/lang/Class;", null, null);
            }
        }
        String bridgeMethodName = methodInfo.name + "$bridge$" + bridgeMethodCounter;
        bridgeMethodCounter++;

        MethodVisitor p = proxyWriter.visitMethod(ACC_PUBLIC | ACC_FINAL, methodInfo.name,
                methodInfo.descriptor, methodInfo.genericSignature, methodInfo.exceptionNames);
        MethodVisitor b = bridgeWriter.visitMethod(ACC_PUBLIC | ACC_STATIC, bridgeMethodName,
                bridgeMethodDescriptor, null, null);
        if (LOG_DEPRECATED && interfaceMethod.isAnnotationPresent(Deprecated.class)) {
            logDeprecated(p, "Warning: using deprecated JBR API method " +
                    interfaceMethod.getDeclaringClass().getName() + "#" + interfaceMethod.getName());
        }
        MethodVisitor bp = generateBridge ? b : p;
        bp.visitFieldInsn(GETSTATIC, bridgeOrProxyName, handleName, MH_DESCRIPTOR);
        if (passInstance) {
            p.visitVarInsn(ALOAD, 0);
            p.visitFieldInsn(GETFIELD, proxyName, "target", OBJECT_DESCRIPTOR);
            b.visitVarInsn(ALOAD, 0);
        }
        int lvIndex = 1;
        for (TypeMapping param : mapping.parameterMapping) {
            int opcode = getLoadOpcode(param.from);
            p.visitVarInsn(opcode, lvIndex);
            b.visitVarInsn(opcode, lvIndex - (passInstance ? 0 : 1));
            lvIndex += getParameterSize(param.from);
            convertValue(bp, bridgeOrProxyName, param);
        }
        if (generateBridge) {
            p.visitMethodInsn(INVOKESTATIC, bridgeName, bridgeMethodName, bridgeMethodDescriptor, false);
        }
        bp.visitMethodInsn(INVOKEVIRTUAL, MH_NAME, "invoke", bridgeMethodDescriptor, false);
        convertValue(bp, bridgeOrProxyName, mapping.returnMapping);
        int returnOpcode = getReturnOpcode(mapping.returnMapping.to);
        p.visitInsn(returnOpcode);
        b.visitInsn(returnOpcode);
        p.visitMaxs(-1, -1);
        b.visitMaxs(-1, -1);
    }

    private String addHandle(ClassVisitor classWriter, Supplier<MethodHandle> handleSupplier) {
        String handleName = "h" + handles.size();
        handles.add(handleSupplier);
        classWriter.visitField(ACC_PRIVATE | ACC_STATIC, handleName, MH_DESCRIPTOR, null, null);
        return handleName;
    }

    private static void convertValue(MethodVisitor m, String handlesHolderName, TypeMapping mapping) {
        if (mapping.conversion == TypeConversion.IDENTITY) return;
        Label skipConvert = new Label();
        m.visitInsn(DUP);
        m.visitJumpInsn(IFNULL, skipConvert);
        switch (mapping.conversion) {
            case EXTRACT_TARGET:
                    m.visitFieldInsn(GETSTATIC, handlesHolderName, mapping.metadata.extractTargetHandle, MH_DESCRIPTOR);
                    break;
            case WRAP_INTO_PROXY:
                    m.visitFieldInsn(GETSTATIC, handlesHolderName, mapping.metadata.proxyConstructorHandle, MH_DESCRIPTOR);
                    break;
            case DYNAMIC_2_WAY:
                m.visitInsn(DUP);
                m.visitMethodInsn(INVOKEVIRTUAL, "java/lang/Object", "getClass", "()Ljava/lang/Class;", false);
                m.visitFieldInsn(GETSTATIC, handlesHolderName, mapping.metadata.extractableClassField, "Ljava/lang/Class;");
                Label elseBranch = new Label(), afterBranch = new Label();
                m.visitJumpInsn(IF_ACMPNE, elseBranch);
                m.visitFieldInsn(GETSTATIC, handlesHolderName, mapping.metadata.extractTargetHandle, MH_DESCRIPTOR);
                m.visitJumpInsn(GOTO, afterBranch);
                m.visitLabel(elseBranch);
                m.visitFieldInsn(GETSTATIC, handlesHolderName, mapping.metadata.proxyConstructorHandle, MH_DESCRIPTOR);
                m.visitLabel(afterBranch);
                break;
        }
        m.visitInsn(SWAP);
        m.visitMethodInsn(INVOKEVIRTUAL, MH_NAME, "invoke", CONVERSION_DESCRIPTOR, false);
        m.visitLabel(skipConvert);
    }

    private static MethodMapping getTargetMethodMapping(Method interfaceMethod) {
        Class<?>[] params = interfaceMethod.getParameterTypes();
        TypeMapping[] paramMappings = new TypeMapping[params.length];
        for (int i = 0; i < params.length; i++) {
            paramMappings[i] = getTargetTypeMapping(params[i]);
            params[i] = paramMappings[i].to;
        }
        TypeMapping returnMapping = getTargetTypeMapping(interfaceMethod.getReturnType()).inverse();
        return new MethodMapping(MethodType.methodType(returnMapping.from, params), returnMapping, paramMappings);
    }

    private static <T> TypeMapping getTargetTypeMapping(Class<T> userType) {
        TypeMappingMetadata m = new TypeMappingMetadata();
        Proxy<T> p = JBRApi.getProxy(userType);
        if (p != null && p.getInfo().target != null) {
            Proxy<?> r = JBRApi.getProxy(p.getInfo().target.lookupClass());
            if (r != null && r.getInfo().target != null) {
                return new TypeMapping(userType, p.getInfo().target.lookupClass(), TypeConversion.DYNAMIC_2_WAY, p, r, m);
            }
            return new TypeMapping(userType, p.getInfo().target.lookupClass(), TypeConversion.EXTRACT_TARGET, p, null, m);
        }
        Class<?> interFace = JBRApi.getProxyInterfaceByTargetName(userType.getName());
        if (interFace != null) {
            Proxy<?> r = JBRApi.getProxy(interFace);
            if (r != null) {
                return new TypeMapping(userType, interFace, TypeConversion.WRAP_INTO_PROXY, null, r, m);
            }
        }
        return new TypeMapping(userType, userType, TypeConversion.IDENTITY, null, null, m);
    }

    private static class MethodMapping implements Iterable<TypeMapping> {
        private final MethodType type;
        private final TypeMapping returnMapping;
        private final TypeMapping[] parameterMapping;
        private MethodMapping(MethodType type, TypeMapping returnMapping, TypeMapping[] parameterMapping) {
            this.type = type;
            this.returnMapping = returnMapping;
            this.parameterMapping = parameterMapping;
        }
        @Override
        public Iterator<TypeMapping> iterator() {
            return new Iterator<>() {
                private int index = -1;
                @Override
                public boolean hasNext() {
                    return index < parameterMapping.length;
                }
                @Override
                public TypeMapping next() {
                    TypeMapping m = index == -1 ? returnMapping : parameterMapping[index];
                    index++;
                    return m;
                }
            };
        }

        /**
         * Every convertable parameter type is replaced with {@link Object} for bridge descriptor.
         * Optional {@link Object} is added as first parameter for instance methods.
         */
        String getBridgeDescriptor(boolean passInstance) {
            StringBuilder bd = new StringBuilder("(");
            if (passInstance) bd.append(OBJECT_DESCRIPTOR);
            for (TypeMapping m : parameterMapping) {
                bd.append(m.getBridgeDescriptor());
            }
            bd.append(')');
            bd.append(returnMapping.getBridgeDescriptor());
            return bd.toString();
        }
    }

    private static class TypeMapping {
        private final Class<?> from, to;
        private final TypeConversion conversion;
        private final Proxy<?> fromProxy, toProxy;
        private final TypeMappingMetadata metadata;
        private TypeMapping(Class<?> from, Class<?> to, TypeConversion conversion,
                Proxy<?> fromProxy, Proxy<?> toProxy, TypeMappingMetadata metadata) {
            this.from = from;
            this.to = to;
            this.conversion = conversion;
            this.fromProxy = fromProxy;
            this.toProxy = toProxy;
            this.metadata = metadata;
        }
        TypeMapping inverse() {
            TypeConversion c;
            switch (conversion) {
                case EXTRACT_TARGET:
                    c = TypeConversion.WRAP_INTO_PROXY;
                    break;
                case WRAP_INTO_PROXY:
                    c = TypeConversion.EXTRACT_TARGET;
                    break;
                default:
                    c = conversion;
                    break;
            }
            return new TypeMapping(to, from, c, toProxy, fromProxy, metadata);
        }
        String getBridgeDescriptor() {
            if (conversion == TypeConversion.IDENTITY) return Type.getDescriptor(from);
            else return "Ljava/lang/Object;";
        }
    }

    private static class TypeMappingMetadata {
        private String extractTargetHandle, proxyConstructorHandle, extractableClassField;
    }

    private enum TypeConversion {
        /**
         * No conversion.
         */
        IDENTITY,
        /**
         * Take a proxy object and extract its target implementation object.
         */
        EXTRACT_TARGET,
        /**
         * Create new proxy targeting given implementation object.
         */
        WRAP_INTO_PROXY,
        /**
         * Decide between {@link #EXTRACT_TARGET} and {@link #WRAP_INTO_PROXY} at runtime, depending on actual object.
         */
        DYNAMIC_2_WAY
    }
}
